package com.chelvc.framework.base.util;

import java.util.Collection;
import java.util.List;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;

/**
 * 线程工具类
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
public final class ThreadUtils {
    private ThreadUtils() {
    }

    /**
     * 休眠线程
     *
     * @param millis 毫秒数
     * @return true/false
     */
    public static boolean sleep(long millis) {
        try {
            Thread.sleep(millis);
        } catch (InterruptedException ignored) {
            return false;
        }
        return true;
    }

    /**
     * 异步调用方法
     *
     * @param runnable 回调接口
     * @return 线程对象
     */
    public static Thread run(@NonNull Runnable runnable) {
        Thread thread = new Thread(runnable);
        thread.start();
        return thread;
    }

    /**
     * 异步调用方法
     *
     * @param runs 回调接口数组
     * @return 线程对象数组
     */
    public static Thread[] run(@NonNull Runnable... runs) {
        if (runs.length == 0) {
            return new Thread[0];
        }
        Thread[] threads = new Thread[runs.length];
        for (int i = 0; i < runs.length; i++) {
            threads[i] = run(runs[i]);
        }
        return threads;
    }

    /**
     * 阻塞线程
     *
     * @param threads 线程对象数组
     */
    public static void join(@NonNull Thread... threads) {
        for (Thread thread : threads) {
            try {
                thread.join();
            } catch (InterruptedException ignored) {
                break;
            }
        }
    }

    /**
     * 等待线程执行完成
     *
     * @param threads 线程对象集合
     */
    public static void join(@NonNull Collection<Thread> threads) {
        for (Thread thread : threads) {
            try {
                thread.join();
            } catch (InterruptedException ignored) {
                break;
            }
        }
    }

    /**
     * 异步消费消息
     *
     * @param queue    阻塞队列
     * @param consumer 消息消费者
     * @param <T>      消息类型
     * @return 消费线程列表
     */
    public static <T> List<Thread> consume(@NonNull BlockingQueue<T> queue, @NonNull Consumer<T> consumer) {
        return consume(queue, consumer, 3);
    }

    /**
     * 异步消费消息
     *
     * @param queue       阻塞队列
     * @param consumer    消息消费者
     * @param concurrency 并发数量
     * @param <T>         消息类型
     * @return 消费线程列表
     */
    public static <T> List<Thread> consume(@NonNull BlockingQueue<T> queue,
                                           @NonNull Consumer<T> consumer, int concurrency) {
        return consume(queue, consumer, () -> false, concurrency);
    }

    /**
     * 异步消费消息
     *
     * @param queue    阻塞队列
     * @param consumer 消息消费者
     * @param breaker  中断方法（返回true表示中断）
     * @param <T>      消息类型
     * @return 消费线程列表
     */
    public static <T> List<Thread> consume(@NonNull BlockingQueue<T> queue,
                                           @NonNull Consumer<T> consumer,
                                           @NonNull Supplier<Boolean> breaker) {
        return consume(queue, consumer, breaker, 3);
    }

    /**
     * 异步消费消息
     *
     * @param queue       阻塞队列
     * @param consumer    消息消费者
     * @param breaker     中断方法（返回true表示中断）
     * @param concurrency 并发数量
     * @param <T>         消息类型
     * @return 消费线程列表
     */
    public static <T> List<Thread> consume(@NonNull BlockingQueue<T> queue, @NonNull Consumer<T> consumer,
                                           @NonNull Supplier<Boolean> breaker, int concurrency) {
        return IntStream.range(0, concurrency).mapToObj(i -> run(() -> {
            while (!Thread.interrupted() && !Boolean.TRUE.equals(breaker.get())) {
                T message;
                try {
                    message = queue.poll(1, TimeUnit.SECONDS);
                } catch (InterruptedException ignored) {
                    break;
                }
                if (message != null) {
                    try {
                        consumer.accept(message);
                    } catch (Exception e) {
                        log.error(e.getMessage(), e);
                    }
                }
            }
        })).collect(Collectors.toList());
    }

    /**
     * 异步消费消息
     *
     * @param consumer 消息消费者
     * @param <T>      消息类型
     * @return 消息队列
     */
    public static <T> BlockingQueue<T> queue(@NonNull Consumer<T> consumer) {
        return queue(consumer, () -> false);
    }

    /**
     * 异步消费消息
     *
     * @param consumer 消息消费者
     * @param breaker  中断方法（返回true表示中断）
     * @param <T>      消息类型
     * @return 消息队列
     */
    public static <T> BlockingQueue<T> queue(@NonNull Consumer<T> consumer, @NonNull Supplier<Boolean> breaker) {
        return queue(1000, consumer, breaker);
    }

    /**
     * 异步消费消息
     *
     * @param capacity 队列大小
     * @param consumer 消息消费者
     * @param <T>      消息类型
     * @return 消息队列
     */
    public static <T> BlockingQueue<T> queue(int capacity, @NonNull Consumer<T> consumer) {
        return queue(capacity, consumer, () -> false);
    }

    /**
     * 异步消费消息
     *
     * @param capacity 队列大小
     * @param consumer 消息消费者
     * @param breaker  中断方法（返回true表示中断）
     * @param <T>      消息类型
     * @return 消息队列
     */
    public static <T> BlockingQueue<T> queue(int capacity, @NonNull Consumer<T> consumer,
                                             @NonNull Supplier<Boolean> breaker) {
        return queue(capacity, consumer, breaker, 3);
    }

    /**
     * 异步消费消息
     *
     * @param consumer    消息消费者
     * @param concurrency 并发数量
     * @param <T>         消息类型
     * @return 消息队列
     */
    public static <T> BlockingQueue<T> queue(@NonNull Consumer<T> consumer, int concurrency) {
        return queue(consumer, () -> false, concurrency);
    }

    /**
     * 异步消费消息
     *
     * @param consumer    消息消费者
     * @param breaker     中断方法（返回true表示中断）
     * @param concurrency 并发数量
     * @param <T>         消息类型
     * @return 消息队列
     */
    public static <T> BlockingQueue<T> queue(@NonNull Consumer<T> consumer, @NonNull Supplier<Boolean> breaker,
                                             int concurrency) {
        return queue(1000, consumer, breaker, concurrency);
    }

    /**
     * 异步消费消息
     *
     * @param capacity    队列大小
     * @param consumer    消息消费者
     * @param concurrency 并发数量
     * @param <T>         消息类型
     * @return 消息队列
     */
    public static <T> BlockingQueue<T> queue(int capacity, @NonNull Consumer<T> consumer, int concurrency) {
        return queue(capacity, consumer, () -> false, concurrency);
    }

    /**
     * 异步消费消息
     *
     * @param capacity    队列大小
     * @param consumer    消息消费者
     * @param breaker     中断方法（返回true表示中断）
     * @param concurrency 并发数量
     * @param <T>         消息类型
     * @return 消息队列
     */
    public static <T> BlockingQueue<T> queue(int capacity, @NonNull Consumer<T> consumer,
                                             @NonNull Supplier<Boolean> breaker, int concurrency) {
        BlockingQueue<T> queue = new ArrayBlockingQueue<>(Math.max(capacity, 1));
        consume(queue, consumer, breaker, concurrency);
        return queue;
    }
}