package com.chelvc.framework.jpush;

import java.lang.reflect.Field;
import java.security.GeneralSecurityException;
import java.security.KeyFactory;
import java.security.PrivateKey;
import java.security.spec.PKCS8EncodedKeySpec;
import java.util.Arrays;
import java.util.Base64;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import javax.annotation.PostConstruct;
import javax.crypto.Cipher;

import cn.jiguang.common.ClientConfig;
import cn.jiguang.common.ServiceHelper;
import cn.jiguang.common.connection.ApacheHttpClient;
import cn.jiguang.common.resp.APIConnectionException;
import cn.jiguang.common.resp.APIRequestException;
import cn.jmessage.api.JMessageClient;
import cn.jmessage.api.common.model.RegisterInfo;
import cn.jmessage.api.user.UserInfoResult;
import cn.jmessage.api.user.UserStateListResult;
import cn.jmessage.api.user.UserStateResult;
import cn.jpush.api.JPushClient;
import cn.jpush.api.push.CIDResult;
import cn.jpush.api.push.PushClient;
import cn.jpush.api.push.PushResult;
import cn.jpush.api.push.model.BatchPushResult;
import cn.jpush.api.push.model.Message;
import cn.jpush.api.push.model.PushPayload;
import cn.jpush.api.push.model.notification.Notification;
import cn.jpush.api.report.ReceivedsResult;
import com.chelvc.framework.base.context.JacksonContextHolder;
import com.chelvc.framework.base.model.Period;
import com.chelvc.framework.base.model.Platform;
import com.chelvc.framework.base.util.DateUtils;
import com.chelvc.framework.base.util.ObjectUtils;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.jpush.config.JPushProperties;
import com.chelvc.framework.jpush.util.JPushUtils;
import com.google.common.collect.ImmutableMap;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;
import org.springframework.web.client.RestTemplate;

/**
 * 极光登录操作默认实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Component
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class DefaultJPushHandler implements JPushHandler {
    /**
     * 极光令牌验证地址
     */
    private static final String TOKEN2MOBILE_URL = "https://api.verification.jpush.cn/v1/web/loginTokenVerify";

    /**
     * 获取消息地址
     */
    private static final String GET_MESSAGE_URL =
            "https://report.im.jpush.cn/v2/users/%s/messages?count=%d&begin_time=%s&end_time=%s";

    /**
     * 获取历史消息地址
     */
    private static final String GET_HISTORY_MESSAGE_URL = "https://report.im.jpush.cn/v2/users/%s/messages?cursor=%s";
    private final RestTemplate restTemplate;
    private final JPushProperties properties;
    @Value("${spring.profiles.active}")
    private String env;
    private PrivateKey privateKey;
    private JPushClient pushClient;
    private JMessageClient messageClient;

    @PostConstruct
    public void initialize() throws GeneralSecurityException {
        // 初始化登录认证配置
        JPushProperties.Login login = this.properties.getLogin();
        if (login != null && StringUtils.nonEmpty(login.getSecret())) {
            // 初始化私钥
            byte[] secret = Base64.getDecoder().decode(login.getSecret());
            PKCS8EncodedKeySpec keySpec = new PKCS8EncodedKeySpec(secret);
            this.privateKey = KeyFactory.getInstance("RSA").generatePrivate(keySpec);
        }

        // 初始化消息推送配置
        JPushProperties.Push push = this.properties.getPush();
        if (push != null && StringUtils.nonEmpty(push.getAppkey()) && StringUtils.nonEmpty(push.getSecret())) {
            // 初始化推送客户端实例
            this.pushClient = new JPushClient(push.getSecret(), push.getAppkey()) {
                {
                    try {
                        Field field = JPushClient.class.getDeclaredField("_pushClient");
                        field.setAccessible(true);
                        field.set(this, new PushClient(push.getSecret(), push.getAppkey()) {
                            @Override
                            public CIDResult getCidList(int count, String type) throws APIConnectionException,
                                    APIRequestException {
                                // 对批量推送CID排序，确保推送结果顺序与发送数据顺序一致
                                CIDResult result = super.getCidList(count, type);
                                if (result != null && !CollectionUtils.isEmpty(result.cidlist)) {
                                    result.cidlist.sort(Comparator.naturalOrder());
                                }
                                return result;
                            }
                        });
                    } catch (NoSuchFieldException | IllegalAccessException e) {
                        throw new RuntimeException(e);
                    }
                }
            };

            // 初始化消息客户端实例
            this.messageClient = new JMessageClient(push.getAppkey(), push.getSecret());
            this.messageClient.setHttpClient(new ApacheHttpClient(
                    ServiceHelper.getBasicAuthorization(push.getAppkey(), push.getSecret()),
                    null,
                    ClientConfig.getInstance()
            ));
        }
    }

    /**
     * 判断是否是生产环境
     *
     * @return true/false
     */
    private boolean isProduction() {
        return Objects.equals(this.env, "prod");
    }

    /**
     * 获取推送客户端实例
     *
     * @return 推送客户端实例
     */
    private JPushClient getPushClient() {
        return Objects.requireNonNull(this.pushClient, "JPush client has not been initialized");
    }

    /**
     * 获取消息客户端实例
     *
     * @return 消息客户端实例
     */
    private JMessageClient getMessageClient() {
        return Objects.requireNonNull(this.messageClient, "JMessage client has not been initialized");
    }

    /**
     * 获取聊天消息
     *
     * @param url 接口地址
     * @return 聊天消息
     */
    private ChatMessage getChatMessage(String url) {
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}", url);
        }
        String appkey = ObjectUtils.ifNull(this.properties.getPush(), JPushProperties.Push::getAppkey);
        String secret = ObjectUtils.ifNull(this.properties.getPush(), JPushProperties.Push::getSecret);
        HttpHeaders headers = new HttpHeaders();
        headers.add("Authorization", ServiceHelper.getBasicAuthorization(appkey, secret));
        headers.setContentType(MediaType.APPLICATION_JSON);
        HttpEntity<?> entity = new HttpEntity<>(headers);
        ChatMessage message = this.restTemplate.exchange(url, HttpMethod.GET, entity, ChatMessage.class).getBody();
        if (debug) {
            log.debug("JPush response: {}", message);
        }
        return message;
    }

    @Override
    public String token2mobile(@NonNull String token) {
        PrivateKey key = Objects.requireNonNull(this.privateKey, "Private key has not been initialized");

        // 根据令牌获取手机号密文
        Map<String, String> parameters = ImmutableMap.of("loginToken", token);
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}, {}", TOKEN2MOBILE_URL, parameters);
        }
        String authorization = ObjectUtils.ifNull(this.properties.getLogin(), JPushProperties.Login::getAuthorization);
        HttpHeaders headers = new HttpHeaders();
        headers.add("Authorization", authorization);
        headers.setContentType(MediaType.APPLICATION_JSON);
        HttpEntity<?> entity = new HttpEntity<>(parameters, headers);
        JPushMobile response =
                this.restTemplate.exchange(TOKEN2MOBILE_URL, HttpMethod.POST, entity, JPushMobile.class).getBody();
        if (debug) {
            log.debug("JPush response: {}", response);
        }
        Objects.requireNonNull(response, "Invalid jpush server response");
        if (response.getCode() == null || response.getCode() != 8000) {
            throw new RuntimeException(String.format("[%d] %s", response.getCode(), response.getContent()));
        }

        // 解码手机号
        try {
            Cipher cipher = Cipher.getInstance("RSA");
            cipher.init(Cipher.DECRYPT_MODE, key);
            return new String(cipher.doFinal(Base64.getDecoder().decode(response.getPhone())));
        } catch (GeneralSecurityException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void register(@NonNull RegisterInfo... registrations) {
        if (registrations.length > 0) {
            boolean debug = log.isDebugEnabled();
            if (debug) {
                log.debug("JPush request: {}", Arrays.toString(registrations));
            }
            String result;
            try {
                result = this.getMessageClient().registerUsers(registrations);
            } catch (APIConnectionException e) {
                throw new RuntimeException(e);
            } catch (APIRequestException e) {
                throw new RuntimeException(e.getErrorMessage(), e);
            }
            if (debug) {
                log.debug("JPush response: {}", result);
            }
        }
    }

    @Override
    public void register(@NonNull Collection<RegisterInfo> registrations) {
        if (!registrations.isEmpty()) {
            this.register(registrations.toArray(new RegisterInfo[0]));
        }
    }

    @Override
    public List<String> getPushIdentities(int count) {
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}", count);
        }
        List<String> identities;
        try {
            identities = this.getPushClient().getCidList(count, "push").cidlist;
        } catch (APIConnectionException e) {
            throw new RuntimeException(e);
        } catch (APIRequestException e) {
            throw new RuntimeException(e.getErrorMessage(), e);
        }
        if (debug) {
            log.debug("JPush response: {}", identities);
        }
        return identities;
    }

    @Override
    public long push(@NonNull PushPayload payload) {
        return this.push(payload, e -> {
            throw new RuntimeException(e.getErrorMessage(), e);
        });
    }

    @Override
    public long push(@NonNull PushPayload payload, @NonNull Consumer<APIRequestException> failure) {
        // 针对生产环境IOS推送，需要设置apnsProduction为true
        if (this.isProduction()) {
            payload.resetOptionsApnsProduction(true);
        }
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}", payload);
        }
        JPushClient client = this.getPushClient();
        try {
            PushResult result = client.sendPush(payload);
            if (debug) {
                log.debug("JPush response: {}", result);
            }
            if (result != null && result.statusCode == 0) {
                return result.msg_id;
            }
            Integer code = result == null || result.error == null ? null : result.error.getCode();
            String message = result == null || result.error == null ? null : result.error.getMessage();
            throw new RuntimeException(String.format("[%d] %s", code, message));
        } catch (APIConnectionException e) {
            throw new RuntimeException(e);
        } catch (APIRequestException e) {
            failure.accept(e);
        }
        return 0;
    }

    @Override
    public long push(@NonNull Platform platform, @NonNull Message message, @NonNull String... aliases) {
        return this.push(JPushUtils.payload(platform, message, aliases));
    }

    @Override
    public long push(@NonNull Platform platform, @NonNull Message message, @NonNull Collection<String> aliases) {
        return this.push(JPushUtils.payload(platform, message, aliases));
    }

    @Override
    public long push(@NonNull Platform platform, @NonNull Notification notification, @NonNull String... aliases) {
        return this.push(JPushUtils.payload(platform, notification, aliases));
    }

    @Override
    public long push(@NonNull Platform platform, @NonNull Notification notification,
                     @NonNull Collection<String> aliases) {
        return this.push(JPushUtils.payload(platform, notification, aliases));
    }

    @Override
    public long push(@NonNull Platform platform, @NonNull String title, @NonNull String content,
                     @NonNull String... aliases) {
        return this.push(platform, title, content, Collections.emptyMap(), aliases);
    }

    @Override
    public long push(@NonNull Platform platform, @NonNull String title, @NonNull String content,
                     @NonNull Collection<String> aliases) {
        return this.push(platform, title, content, Collections.emptyMap(), aliases);
    }

    @Override
    public long push(@NonNull Platform platform, @NonNull String title, @NonNull String content,
                     @NonNull Map<String, String> extras, @NonNull String... aliases) {
        Notification notification = JPushUtils.notification(platform, title, content, extras);
        return this.push(platform, notification, aliases);
    }

    @Override
    public long push(@NonNull Platform platform, @NonNull String title, @NonNull String content,
                     @NonNull Map<String, String> extras, @NonNull Collection<String> aliases) {
        Notification notification = JPushUtils.notification(platform, title, content, extras);
        return this.push(platform, notification, aliases);
    }

    @Override
    public BatchPushResult push(@NonNull PushPayload... payloads) {
        Assert.isTrue(payloads.length > 0, "payloads must not be empty");
        return this.push(Arrays.asList(payloads));
    }

    @Override
    public BatchPushResult push(@NonNull List<PushPayload> payloads) {
        Assert.isTrue(!payloads.isEmpty(), "payloads must not be empty");
        // 针对生产环境IOS推送，需要设置apnsProduction为true
        if (this.isProduction()) {
            payloads.forEach(payload -> payload.resetOptionsApnsProduction(true));
        }
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}", payloads);
        }

        // 批量推送消息
        BatchPushResult result;
        JPushClient client = this.getPushClient();
        try {
            result = client.batchSendPushByAlias(payloads);
        } catch (APIConnectionException e) {
            throw new RuntimeException(e);
        } catch (APIRequestException e) {
            throw new RuntimeException(e.getErrorMessage(), e);
        }
        if (debug) {
            log.debug("JPush response: {}", JacksonContextHolder.serialize(result));
        }
        return result;
    }

    /**
     * 获取消息投递结果
     *
     * @param ids 消息ID（多个ID之间使用","号隔开，最多最多支持 100 个消息ID）
     * @return 投递结果列表
     */
    private List<ReceivedsResult.Received> getDeliveries(String ids) {
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}", ids);
        }
        if (StringUtils.isEmpty(ids)) {
            return Collections.emptyList();
        }
        ReceivedsResult result;
        JPushClient client = this.getPushClient();
        try {
            result = client.getReceivedsDetail(ids);
        } catch (APIConnectionException e) {
            throw new RuntimeException(e);
        } catch (APIRequestException e) {
            throw new RuntimeException(e.getErrorMessage(), e);
        }
        if (debug) {
            log.debug("JPush response: {}", JacksonContextHolder.serialize(result));
        }
        return ObjectUtils.ifNull(result.received_list, Collections::emptyList);
    }

    @Override
    public ReceivedsResult.Received getDelivery(@NonNull Long id) {
        List<ReceivedsResult.Received> deliveries = this.getDeliveries(String.valueOf(id));
        return CollectionUtils.isEmpty(deliveries) ? null : deliveries.get(0);
    }

    @Override
    public List<ReceivedsResult.Received> getDeliveries(@NonNull Long... ids) {
        return ids.length == 0 ? Collections.emptyList() : this.getDeliveries(Arrays.asList(ids));
    }

    @Override
    public List<ReceivedsResult.Received> getDeliveries(@NonNull Collection<Long> ids) {
        return this.getDeliveries(
                ids.stream().filter(Objects::nonNull).map(Object::toString).collect(Collectors.joining(","))
        );
    }

    @Override
    public UserInfoResult getUserInfo(@NonNull String username) {
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}", username);
        }
        UserInfoResult user;
        try {
            user = this.getMessageClient().getUserInfo(username);
        } catch (APIConnectionException e) {
            throw new RuntimeException(e);
        } catch (APIRequestException e) {
            throw new RuntimeException(e.getErrorMessage(), e);
        }
        if (debug) {
            log.debug("JPush response: {}", user);
        }
        return user;
    }

    @Override
    public UserStateResult getUserState(@NonNull String username) {
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}", username);
        }
        UserStateResult state;
        try {
            state = this.getMessageClient().getUserState(username);
        } catch (APIConnectionException e) {
            throw new RuntimeException(e);
        } catch (APIRequestException e) {
            throw new RuntimeException(e.getErrorMessage(), e);
        }
        if (debug) {
            log.debug("JPush response: {}", state);
        }
        return state;
    }

    @Override
    public UserStateListResult[] getUserStates(@NonNull String... usernames) {
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("JPush request: {}", Arrays.toString(usernames));
        }
        if (usernames.length == 0) {
            return new UserStateListResult[0];
        }
        UserStateListResult[] states;
        try {
            states = this.getMessageClient().getUsersState(usernames);
        } catch (APIConnectionException e) {
            throw new RuntimeException(e);
        } catch (APIRequestException e) {
            throw new RuntimeException(e.getErrorMessage(), e);
        }
        if (debug) {
            log.debug("JPush response: {}", Arrays.toString(states));
        }
        return states;
    }

    @Override
    public ChatMessage getUserMessage(@NonNull String username, @NonNull Period period, int count) {
        return this.getChatMessage(String.format(
                GET_MESSAGE_URL, username, count, DateUtils.format(period.getBegin()), DateUtils.format(period.getEnd())
        ));
    }

    @Override
    public ChatMessage getUserHistoryMessage(@NonNull String username, @NonNull String cursor) {
        return this.getChatMessage(String.format(GET_HISTORY_MESSAGE_URL, username, cursor));
    }
}
