package com.chelvc.framework.security.context;

import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.TreeMap;
import java.util.regex.Pattern;
import javax.servlet.ServletRequestEvent;
import javax.servlet.ServletRequestListener;

import com.beust.jcommander.internal.Lists;
import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.JacksonContextHolder;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.util.StringUtils;
import com.chelvc.framework.security.config.SecurityProperties;
import com.chelvc.framework.security.util.TokenUtils;
import com.fasterxml.jackson.core.type.TypeReference;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.digest.HmacAlgorithms;
import org.apache.commons.codec.digest.HmacUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.annotation.Configuration;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.util.CollectionUtils;

/**
 * 安全处理上下文工具类
 * <p>
 * 角色资源授权信息从Nacos配置中心获取并动态更新
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
@Configuration
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class SecurityContextHolder implements ApplicationContextAware, ServletRequestListener {
    /**
     * 权限属性名称
     */
    public static final String PERMISSION_PROPERTY = "platform.permission";

    /**
     * 方法资源拆分匹配模式
     */
    private static final Pattern METHOD_RESOURCE_SPLIT_PATTERN = StringUtils.getPattern(" +");

    /**
     * 权限类型引用
     */
    private static final TypeReference<Map<String, Set<String>>> PERMISSION_TYPE_REFERENCE =
            new TypeReference<Map<String, Set<String>>>() {
            };

    /**
     * JWT上下文
     */
    private static final ThreadLocal<Jwt> JWT_CONTEXT = new ThreadLocal<>();

    /**
     * 数字签名工具类实例
     */
    private static HmacUtils HMAC_UTILS = null;

    /**
     * 安全配置属性
     */
    private final SecurityProperties securityProperties;

    /**
     * 获取JWT信息
     *
     * @return JWT信息
     */
    public static Jwt getJwt() {
        return getJwt(false);
    }

    /**
     * 设置JWT信息
     *
     * @param jwt JWT信息
     */
    public static void setJwt(Jwt jwt) {
        JWT_CONTEXT.set(jwt);
    }

    /**
     * 获取JWT信息
     *
     * @param nullable JWT是否可以为空
     * @return JWT信息
     */
    public static Jwt getJwt(boolean nullable) {
        Jwt jwt = JWT_CONTEXT.get();
        return nullable ? jwt : Objects.requireNonNull(jwt, "Jwt has not been initialized");
    }

    /**
     * 获取上下文中角色/权限映射表
     *
     * @return 角色/权限映射表
     */
    public static Map<String, Set<String>> getContextPermissions() {
        String permission = ApplicationContextHolder.getProperty(PERMISSION_PROPERTY);
        return StringUtils.isEmpty(permission) ? Collections.emptyMap() :
                JacksonContextHolder.deserialize(permission, PERMISSION_TYPE_REFERENCE);
    }

    /**
     * 判断当前用户是否拥有指定资源权限
     *
     * @param method 请求方式
     * @param uri    资源地址
     * @return true/false
     */
    public static boolean hasPermission(String method, String uri) {
        if (StringUtils.isEmpty(uri)) {
            return false;
        }

        // 基于用户角色及对应的资源判断是否拥有权限
        Set<String> roles = TokenUtils.getAuthorities(getJwt(true));
        if (CollectionUtils.isEmpty(roles)) {
            return false;
        }
        Map<String, Set<String>> permissions = getContextPermissions();
        if (CollectionUtils.isEmpty(permissions)) {
            return false;
        }
        return roles.stream().anyMatch(role -> {
            Set<String> resources = permissions.get(role);
            return !CollectionUtils.isEmpty(resources) && resources.stream().anyMatch(resource -> {
                if (StringUtils.isEmpty(resource) || (resource = resource.trim()).isEmpty()) {
                    return false;
                }
                String[] sections = METHOD_RESOURCE_SPLIT_PATTERN.split(resource);
                if (sections.length == 1) {
                    return StringUtils.isPath(sections[0], uri);
                }
                return Objects.equals(method, sections[0]) && StringUtils.isPath(sections[1], uri);
            });
        });
    }

    /**
     * 获取数字签名工具类实例
     *
     * @return 数字签名工具类实例
     */
    public static HmacUtils getHmacUtils() {
        return Objects.requireNonNull(HMAC_UTILS, "Hmac utils has not been initialized");
    }

    /**
     * 数字签名
     *
     * @param parameters 签名参数
     * @return 签名信息
     */
    public static String sign(Map<?, ?> parameters) {
        Session session = SessionContextHolder.getSession();
        StringBuilder plaintext = new StringBuilder().append(session.getTenant()).append(session.getTerminal())
                .append(session.getId());
        if (parameters instanceof TreeMap) {
            parameters.forEach((key, value) -> plaintext.append(key).append(value));
        } else if (!CollectionUtils.isEmpty(parameters)) {
            List<?> keys = Lists.newArrayList(parameters.keySet());
            keys.sort(Comparator.comparing(Object::toString));
            keys.forEach(key -> plaintext.append(key).append(parameters.get(key)));
        }
        return getHmacUtils().hmacHex(plaintext.toString()) + System.currentTimeMillis();
    }

    /**
     * 验证数字签名
     *
     * @param signature  数字签名信息
     * @param parameters 签名参数
     * @return true/false
     */
    public static boolean validateSignature(String signature, Map<?, ?> parameters) {
        return !StringUtils.isEmpty(signature)
                && Objects.equals(signature.substring(0, 64), sign(parameters).substring(0, 64));
    }

    /**
     * 清空上下文信息
     */
    public static void clearSecurityContext() {
        JWT_CONTEXT.remove();
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        HMAC_UTILS = new HmacUtils(HmacAlgorithms.HMAC_SHA_256, this.securityProperties.getSignature().getSecret());
    }

    @Override
    public void requestDestroyed(ServletRequestEvent event) {
        clearSecurityContext();
    }
}
